// Copyright (c) 2017 TeraByte, Inc.  All Rights Reserved.
//
// Disclaimer:
//
// This script is provided AS IS without warranty of any kind. TeraByte, Inc.
// further disclaims all implied warranties including, without limitation,
// any implied warranties of merchantability or of fitness for a particular
// purpose. The entire risk arising out of the use or performance of the
// sample scripts and documentation remains with you. In no event shall
// TeraByte, Inc., its authors, or anyone else involved in the creation,
// production, or delivery of the scripts be liable for any damages whatsoever
// (including, without limitation, damages for loss of business profits,
// business interruption, loss of business information, or other pecuniary
// loss) arising out of the use of or inability to use the sample scripts or
// documentation, even if TeraByte, Inc. has been advised of the possibility
// of such damages.
//
// Permission to use, copy, modify, and distribute this script for use with
// TBScript with or without fee is hereby granted, provided that the above
// copyright notice, disclaimer, this permission notice, and the below script
// engine notice appear in all copies, and that the name of TeraByte Unlimited
// or TeraByte, Inc. not be used in advertising or publicity pertaining to
// distribution of the script without specific, written prior permission.
//
// Script Engine:
//
// This script requires a fully licensed copy of the TeraByte OS Deployment
// Tool Suite (TBOSDTS) which can be purchased directly from the TeraByte
// Unlimited Website at http://www.terabyteunlimited.com/purchase-tbosdt.htm.
// Your use of TBOSDTS is governed by the terms and conditions of the license
// agreement included with the TBOSDTS software.
//
// Purpose:
//
// This include script is provided to help users get started writing their
// own scripts using TBScript.   It includes several useful subroutines 
// described below:
//
//-------------------------------------------------------------------------
// DrawTitle(pause) - Clear screen and draw the title on the screen
//                Input:   pause   - [i] 1 to pause before clearing screen
//-------------------------------------------------------------------------
// EnterNormalTx() - Enter in normal text screen for output of progress
//-------------------------------------------------------------------------
// GetListSelection(title, list, strpos, msgtype) - Get selection of a list of items
//                Input:   title   - [i] title for window
//                         list    - [i] list of items
//                         strpos  - [i] starting selected position in list (.top)
//                         msgtype - [i] type of message ("", "err", "warn")
//                Output:  1 based index of the selected item or 0 if user escaped
//-------------------------------------------------------------------------
// BuildListIndex(list) - Build an index to the list and obtain the number of lines and widest line
//                Input:   list    - [i] list of items
//                Output:  returns an array of line information
//-------------------------------------------------------------------------
// DrawHelpLine(text, h, ti) - Draw a line with a text
//                Input:   text   - [i] text to display
//                         h      - [i] Heigth of the Window above
//                         ti     - [i] text screen information to use
//                Output:  returns with .x and .y set to final x, y location, and .block set to the previous block
//-------------------------------------------------------------------------
// DrawWindow(title, x, y, width, height, ti) - Draw an emptry window to the screen
//                Input:   title  - [i] title for window
//                         x      - [i] x position (-1 to center)
//                         y      - [i] y position (-1 to center)
//                         width  - [i] window width
//                         height - [i] window height
//                         ti     - [i] text screen information to use
//                Output:  returns with .x and .y set to final x, y location
//-------------------------------------------------------------------------
// GetManualText(title, object, oldval) - Draw a window in the middle of the screen and ask the user to enter a text
//                Input:   title   - [i] title of the window to create
//                         object  - [i] name of the object to ask
//                         oldval  - [i] Actual value of object (if needed)
//                Output:  Text entered by user
//-------------------------------------------------------------------------
// GetWindowsSelection(VirtWinDrive) - Get selection of a windows installation
//                Input:   VirtWinDrive  - [i] if accessing a Virtual drive, full path to that virtual drive
//                Output:  TRUE - Windows installation chosen (with .DrvLtr, .hd, .id, .name, .size, .desc)
//-------------------------------------------------------------------------
// GetFileSelection(title, mask, path) - Get selection of a file. The hidden and "system protected" files and directories are not shown to user, and thus, cannot be selected.
//                Input:   title   - [i] title for window
//                         mask    - [i] search mask (don't include path - empty to explore all)
//                         path    - [i] default path
//                Output:  full path of file selected
//-------------------------------------------------------------------------
// BuildPath(path, file) - Combine a path and file depending on the environment (LINUX or DOS/WINDOWS)
//                Input:   path - file path
//                         file - file name
//                Output:  combined path and file
//-------------------------------------------------------------------------
// RemoveComponent(filepath) - Remove component of a file path
//                Input:   filepath  - file path to process
//                Output:  truncated file path
//-------------------------------------------------------------------------
// GetComponent(filepath) - Read file component of file path
//                Input:   filepath  - file path to process
//                Output:  truncated file name
//-------------------------------------------------------------------------
// SplitPath(filepath) - Split a file in to its path / file components
//                Input:   filepath  - file path to process
//                Output:  split path with members .path and .file set
//-------------------------------------------------------------------------
// GetFileExtension(filename) - Extract the extension of a file
//                Input:   filename  - file name to process
//                Output:  file extension without the dot
//-------------------------------------------------------------------------
// Replace(s, char1, char2) - Replace a char inside a string to another char
//                Input:   s      -     String to process
//                         char1  - [i] char to be replaced
//                         char2  - [i] char to replace
//                Output:  s with char1 replaced by char2
//-------------------------------------------------------------------------
// Trim(s) - Remove leading and trailing spaces, tabs, newlines and all non-readable char from a string
//                Input:   s  - string to process
//                Output:  Trim string
//-------------------------------------------------------------------------
// OpenDiskCache(hdnum) - Opens disk cache on a drive
//                Input:   hdnum   - [i] Number of hd
//-------------------------------------------------------------------------
// CloseDiskCache(hdnum, flush) - Closes disk cache on a drive
//                Input:   hdnum   - [i] number of HD
//                         flush   - [i] 1 to flush cache only, 0 to close cache
//-------------------------------------------------------------------------
// Size2Text(size, i) - Evaluate a size in numeric, to a size with unit (B, KiB, MiB, GiB, TiB)
//                Input:   size   -     size
//                         i      - [i] 0 if the entry size is in bytes, 1 if it's KiB, 2 if it's MiB...
//                Output:  Size in B, KiB, MiB, GiB or TiB, rounded at 2 digits after coma

// color schemes
const COLOR_NORMAL         = 0x07
const COLOR_LISTITEM_NORM  = 0x3F
const COLOR_LISTITEM_SEL   = 0x2E   // Normal selection
const COLOR_LISTITEM_ERR   = 0xCF   // Error messages
const COLOR_LISTITEM_WARN  = 0xE1   // Warning messages
const COLOR_WINDOW         = 0x1E
const COLOR_WINDOWTEXT     = 0x1F
const COLOR_INPUT          = 0x2F
const COLOR_TITLE          = 0x1F

// key codes
const KEY_ENTER   = 13
const KEY_ESCAPE  = 27
const KEY_DOWN    = 20480
const KEY_UP      = 18432
const KEY_PGDN    = 20736
const KEY_PGUP    = 18688
const KEY_HOME    = 18176
const KEY_END     = 20224
const KEY_RESIZE  = 410

// file attributes
const FILEATTR_HID = 0x02
const FILEATTR_SYS = 0x04
const FILEATTR_VOL = 0x08
const FILEATTR_DIR = 0x10

// global variables
global gLastPath = ""             // Last Path used when selecting file
global gDrive = 0                 // Array for description of drives mounted by script
const WINDIR = "\Windows"
global gUseHDCache = 1             // set to 1 to enable disk cache use; 0 = disable
const HDLAZY_WRITE   = 1           // set to 1 to enable partial lazy writing; 0 = disable
const HDCACHE_SIZE   = 10*1048576  // 10MiB
const HDCACHE_IOSIZE = 64*1024     // 64KiB IO Cache size

//-------------------------------------------------------------------------
// Purpose: Clear screen and draw the title on the screen
//
// Input:   pause   - [i] if set to 1 will pause before clearing screen
//
// Output:  na
//
// Notes:   Screen is not cleared under complete automatic mode.
//
sub DrawTitle(pause)
	if pause=1 then
		// pause for user to check status
		print("Press <Enter> to continue...")
		while GetKey()<>KEY_ENTER
		wend
		if tbsenv="LINUX" then
			printl("")
		end if
	end if

	// Init Tx mode for screen control
	TxInit()
	TxAscii(1)
	TXSetAttr(COLOR_NORMAL)
	cls()
	ti=TXGetInfo()
	TXGotoXY(1, 1)
	TXSetAttr(COLOR_TITLE)
	print(PAD(SCRIPT_NAME, ti.ViewWidth, 1))
end sub


//-------------------------------------------------------------------------
// Purpose: Enter in normal text screen for output of progress
//
// Input:   na
//
// Output:  na
//
// Notes:
//
sub EnterNormalTx()
	// Move cursor to the beginning of Window
	TXGotoXY(1,3)
	if tbsenv="LINUX" then
		// when Tx mode is active, linux TTY output (internal TBOSDT output) won't
		// display on the screen until the script issues print or printl.  In addtion,
		// newline output will not include cariage returns.  Therefore, we revert back
		// to standard TTY mode by terminating Tx mode.
		TxTerm()
	else
		TXSetAttr(COLOR_NORMAL)
	end if
end sub


//-------------------------------------------------------------------------
// Purpose: Get selection of a list of items
//
// Input:   title   - [i] title for window
//          list    - [i] list of items
//          strpos  - [i] starting selected position in list (.top)
//          msgtype - [i] type of message:
//                      - "err" to display an error message
//                      - "warn" to display a warning message
//                      - or empty to display normal message
//
// Output:  1 based index of the selected item or 0 if user escaped and
//          -1 if view area too small.  Also returns member .top to
//          indicate condition of listbox so caller can reposition if
//          dialog needs to be displayed again.
//
// Notes:   each line item in the input list is terminated using a newline.
//
sub GetListSelection(title, list, strpos, msgtype)
	// get count of list items and widest line
	lineinfo=BuildListIndex(list)

	// initialize controlling variables
	ptr=strpos
	top=strpos.top
	prefix=(lineinfo.count>1)

	// make sure users parameters within range
	if ptr<1 or ptr>lineinfo.count then
		ptr=1
		top=1
	elseif top<1 or top>lineinfo.count then
		top=1
	end if

	resize=1

	while resize<>0
		resize=0

		// get information about current screen
		ti=TXGetInfo()

		// make sure it's something reasonable
		if WaitForScreenSize(15, 8)=0 then
			return -1
		end if

		// disable cursor
		TXCursorType(0)

		// calculate the size of the dialog/list box we want to use
		lbwidth=lineinfo.widest+2*prefix
		lbheight=lineinfo.count

		if len(title,0)+2>lbwidth then
			lbwidth=len(title, 0)+2
		end if

		dlgwidth=lbwidth+4
		dlgheight=lbheight+2

		// make sure it fits the screen
		if dlgwidth>ti.ViewWidth-2 then
			dlgwidth=ti.ViewWidth-2
			lbwidth=dlgwidth-4
		end if

		if dlgheight>ti.ViewHeight-2 then
			dlgheight=ti.ViewHeight-2
			lbheight=dlgheight-2
		end if

		// make sure top within range of ptr user provided
		if top>ptr or top+lbheight<=ptr then
			// put selected item in about the middle
			top=ptr-(lbheight/2)
			// ensure not too low
			if top>=lineinfo.count-lbheight+1 then
				top=lineinfo.count-lbheight+2
			end if
		end if

		// build the main window
		winpos=DrawWindow(title, -1, -1, dlgwidth, dlgheight, ti)

		// initialize tracking variables
		orgtop=0
		orgptr=0

		// Color of the seleted item
		if msgtype="err" then
			colsel=COLOR_LISTITEM_ERR
		elseif msgtype="warn" then
			colsel=COLOR_LISTITEM_WARN
		else
			colsel=COLOR_LISTITEM_SEL
		end if

		// keep looping until we manually exit below
		while 1
			// determine if we have to redraw entire list box
			if top<>orgtop then
				// draw entire list box to screen
				for count=0 to lbheight-1
					TXGotoXY(winpos.x+2, winpos.y+1+count)
					if top+count=ptr then
						TXSetAttr(colsel)
					else
						TXSetAttr(COLOR_LISTITEM_NORM)
					end if
					print(BuildDispText(lineinfo[top+count], lbwidth, prefix*(1+(top+count=ptr))))
				next
			elseif ptr<>orgptr then
				// draw the org line in normal color
				TXGotoXY(winpos.x+2, winpos.y+1+(orgptr-top))
				TXSetAttr(COLOR_LISTITEM_NORM)
				print(BuildDispText(lineinfo[orgptr], lbwidth, prefix))
				// draw the new selected line in selection colors
				TXGotoXY(winpos.x+2, winpos.y+1+(ptr-top))
				TXSetAttr(colsel)
				print(BuildDispText(lineinfo[ptr], lbwidth, 2*prefix))
			end if

			// get user input
			key=GetKey()

			// save current position
			orgptr=ptr
			orgtop=top

			// determine what to do
			if key=KEY_ENTER then
				if lineinfo.count>0 and Left(lineinfo[ptr], 3)<>"--[" then
					ExitLoop
				end if
			elseif key=KEY_ESCAPE then
				ptr=0
				ExitLoop
			elseif key=KEY_UP then
				// check if at top of list box
				if ptr=top then
					// at top of list box ensure something to scroll
					if top>1 then
						// cause scrolling
						top=top-1
						ptr=ptr-1
					end if
				else
					// move line selection up one
					ptr=ptr-1
				end if
			elseif key=KEY_DOWN then
				// ensure not at last available line
				if ptr<lineinfo.count then
					// check if at end of list box
					if ptr=top+lbheight-1 then
						// cause scrolling
						top=top+1
						ptr=ptr+1
					else
						// move line selection down one
						ptr=ptr+1
					end if
				end if
			elseif key=KEY_PGUP then
				// move up an entire page
				ptr=ptr-lbheight
				top=top-lbheight
				// ensure new ptr within range
				if ptr<1 then
					ptr=1
				end if
				// ensure new top within range
				if top<1 then
					top=1
				end if
			elseif key=KEY_PGDN then
				// move down an entire page
				ptr=ptr+lbheight
				top=top+lbheight
				// ensure new ptr within range
				if ptr>lineinfo.count then
					ptr=lineinfo.count
				end if
				// ensure new top within range
				if top>lineinfo.count then
					top=orgtop
				end if
			elseif key=KEY_HOME then
				// all the way to the beginning
				ptr=1
				top=1
			elseif key=KEY_END then
				// all the way to tne end
				ptr=lineinfo.count
				// check if we need to rebuild list box
				if top+lbheight<=lineinfo.count then
					top=lineinfo.count-lbheight+1
				end if
			elseif key=KEY_RESIZE then
				// adjust size of list box
				resize=1
				ExitLoop
			end if
		wend

		// restore screen
		TXPutBlock(winpos.block, winpos.x, winpos.y)
		winpos=""

		// enable standard cursor
		TXCursorType(2)
	wend

	// return position to caller
	ptr.top=top
	return ptr
end sub

//-------------------------------------------------------------------------
// Purpose: Build the text to print on a selection windows
//
// Input:   text    -     text to print
//          l       - [i] length of the text
//          prefix  - [i] 2 if a prefix is to be added, 1 if two spaces have to be added
//
// Output:  Formated text to print
//
// Notes:   Used to center titles
//
sub BuildDispText(text, l, prefix)
	if prefix=2 then
		r="> "
	elseif prefix=1 then
		r="  "
	end if

	if Left(text, 3)<>"--[" then
		// default
		r=PAD(Left(r # text, l, 0), l, 0, 0)
	else
		// title or separator line
		l=l-len(r)
		r=r # PAD(Left(text, l), l, (Right(text, 3)="]--"))
		r=Replace(r, " ", "-")
	end if

	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Build an index to the list and obtain the number of lines
//          and widest line
//
// Input:   list    - [i] list of items
//
// Output:  returns an array of line information as follows:
//             .count  = total lines
//             .widest = widest line
//
// Notes:   each line item in the input list is terminated using a newline
//
sub BuildListIndex(list)
	// Add newline caracter at the end
	list=list # "^n"
	// initialize return value
	r=""
	r.count=0
	r.widest=0

	// start processing each line
	width=InStr(list, "^n")-1
	while width>0
		// found non-empty line
		r.count=r.count+1
		// determine if it's the widest
		w=InStr(list, "^n", 0)-1
		if w>r.widest then
			r.widest=w
		end if
		// save line information
		r[r.count]=Left(list, width)
		// next line
		list=Mid(list, width+2)
		width=InStr(list, "^n")-1
	wend

	// return result to caller
	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Draw a line
//
// Input:   text   - [i] text to display
//          h      - [i] Heigth of the Window above
//          ti     - [i] text screen information to use
//
// Output:  returns with .x and .y set to final x, y location, and .block set to the previous block
//
// Notes:   
//
sub DrawHelpLine(text, h, ti)
	if ti="" then
		ti=TXGetInfo()
	end if

	if text="" then
		text="( Leave blank and press <Enter> to cancel )"
	end if
	text="  " # text # "  "
	// determine width of line to display
	width=len(text)
	if width>=ti.ViewWidth then
		width=ti.ViewWidth-1
	end if

	// calculate position of text to display
	x=((ti.ViewWidth-width)/2)+1
	y=(ti.ViewHeight-h)/2+1+h
	if y>ti.ViewHeight then
		y=ti.ViewHeight
	end if

	// Save result
	r.x=x
	r.y=y
	r.block=TXGetBlock(x, y, x+width-1, y)

	// print the text
	TXGotoXY(x, y)
	TXSetAttr(COLOR_WINDOWTEXT)
	print(PAD(Left(text, width), width, 1))

	// return the result
	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Draw an emptry window to the screen
//
// Input:   title  - [i] title for window
//          x      - [i] x position (-1 to center)
//          y      - [i] y position (-1 to center)
//          width  - [i] window width
//          height - [i] window height
//          ti     - [i] text screen information to use
//
// Output:  returns with .x and .y set to final x, y location
//          (.x and .y will be zero if screen size not large enough)
//
// Notes:
//
sub DrawWindow(title, x, y, width, height, ti)
	// make sure screen size large enough for window
	if WaitForScreenSize(width, height)=0 then
		return
	end if

	// check if ti supplied
	if ti="" then
		ti=TXGetInfo()
	end if

	// limit title to width
	if len(title, 0)>width-4 then
		title=Left(title, width-4, 0)
	end if

	// check if x or y needs to be centered
	if x=-1 then
		x=((ti.ViewWidth-width)/2)+1
	end if

	if y=-1 then
		y=((ti.ViewHeight-height)/2)+1
	end if

	// save result
	r.x=x
	r.y=y
	r.block=TXGetBlock(x, y, x+width-1, y+height-1)

	// setup proper box chars for each target
	if tbsenv="LINUX" then
		box.ul=-108
		box.h=-113
		box.ur=-107
		box.ll=-109
		box.lr=-106
		box.v=-120
	else
		box.ul=201
		box.h=205
		box.ur=187
		box.ll=200
		box.lr=188
		box.v=186
	end if
	box.fill=32

	// create line to buffer for faster output of each line
	TXSetAttr(COLOR_WINDOW)

	// top line
	TXGotoXY(x,y)
	TXOutCh(box.ul)
	TXOutCh(box.h, width-2)
	TXOutCh(box.ur)
	y=y+1

	// middle lines
	for i=3 to height
		TXGotoXY(x,y)
		TXOutCh(box.v)
		TXOutCh(box.fill,width-2)
		TXOutCh(box.v)
		y=y+1
	next

	// bottom line
	TXGotoXY(x,y)
	TXOutCh(box.ll)
	TXOutCh(box.h, width-2)
	TXOutCh(box.lr)

	// force screen update and update title
	if title<>"" then
		title="[ " # title # " ]"
		TXGotoXY(r.x+((width-len(title,0))/2), r.y)
		print(title)
	end if

	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Wait for Screen Size to be resonable
//
// Input:   minwidth   - [i] minimum width required
//          minheight  - [i] minimum height required
//
// Output:  TRUE - set resonable
//          FALSE - user pressed escape
//
// Notes:
//
sub WaitForScreenSize(minwidth, minheight)
	result=1
	ti=TXGetInfo()

	// notify user if too small
	if ti.ViewWidth<minwidth or ti.ViewHeight<minheight then
		TXSetAttr(COLOR_NORMAL)
		cls()
		print("Screen too small")
		// wait for size to change or user to press escape
		while ti.ViewWidth<minwidth or ti.ViewHeight<minheight
			key=GetKey("", 1)
			if key=KEY_ESCAPE then
				result=0
				ExitLoop
			elseif key=KEY_RESIZE then
				// refresh screen
				print("")
			end if
			ti=TXGetInfo()
		wend
		cls()
	end if

	return result
end sub

//-------------------------------------------------------------------------
// Purpose: Draw a window in the middle of the screen and ask the user to enter
//          a text.
//
// Input:   title   - [i] title of the window to create
//          object  - [i] name of the object to ask
//          oldval  - [i] Actual value of object (if needed)
//
// Output:  Text entered by user
//
// Notes:
//
sub GetManualText(title, object, oldval)
	// Init
	text=""

	// make a fake window
	w=len(object)+62+2*(oldval<>"")
	l=5+(oldval<>"")
	winpos=DrawWindow(title, -1, -1, w, l, "")

	if winpos.x>0 then
		linepos=DrawHelpLine("", l, "")
		TXGotoXY(winpos.x+2, winpos.y+2)
		TXSetAttr(COLOR_WINDOWTEXT)
		if oldval<>"" then
			print("Default ", object, ": ")
			TXSetAttr(COLOR_INPUT)
			print(Left(oldval, 50))
			TXGotoXY(winpos.x+4, winpos.y+3)
			TXSetAttr(COLOR_WINDOWTEXT)
		end if
		print("Enter ", object, ": ")
		if object<>"Password" then
			TXSetAttr(COLOR_INPUT)
		else
			TXSetAttr(0x22)
		end if
		// Ask user to enter text
		text=Trim(GetStr("", 50))
		// remove window we just create
		TXPutBlock(winpos.block, winpos.x, winpos.y)
		winpos=""
		if linepos.x>0 then
			// remove line we just create
			TXPutBlock(linepos.block, linepos.x, linepos.y)
			linepos=""
		end if
	end if

	// Return entry to caller
	return text
end sub

//-------------------------------------------------------------------------
// Purpose: Get selection of a windows installation
//
// Input:   VirtWinDrive   - [i] if accessing a Virtual drive, path to that virtual drive
//
// Output:  TRUE - Windows installation chosen (with .DrvLtr, .hd, .id, .name, .size, .desc)
//          FALSE - No windows installation chosen
//
// Notes:   Mount drive 0: if required
//          Open the system registry as registry object 0, and software registry as registry object 1
//          the const WINDIR must be set to the Windows directory (eg. "\Windows")
//
sub GetWindowsSelection(VirtWinDrive)

	// disable various warning / error messages
	ext("set option quiet 1")

	// make sure not already mounted
	ext("close reg 0")
	ext("close reg 1")
	ext("umount 0:")

	// ensure no locking option
	ext("set option locking 0")
	// enable drive letters
	GetDrvLtrInfo(1)

	// start search for windows installation
	listbuf=""
	lineinfo=""
	count=0

	for hd=0 to 63
		// determine if using virtual drive or physical drive search
		if VirtWinDrive<>"" then
			hd=VirtWinDrive
		end if
		h=GetHdInfo(hd)
		if InStr(hd, " ") then
			hd="^"" # hd # "^""
		end if

		if h then
			pn=1
			while h.partition[pn]
				// attempt to mount the partition
				if ext("mount 0: " # hd # " " # h.partition[pn].id)=0 then
					if ext("open reg 0 0:" # WINDIR # "\system32\config\SYSTEM -f")=0 then
						// Close the registry
						ext("close reg 0")
						// save information
						count=count+1
						lineinfo[count]=BuildWinPartInfo(hd, h.partition[pn], h.secsize)
						// add it to the list
						listbuf=listbuf # lineinfo[count].desc # "^n"
					end if
					// close the drive
					ext("umount 0:")
				end if

				// check if it has volumes
				vn=1
				while h.partition[pn].volume[vn]
					if ext("mount 0: " # hd # " " # h.partition[pn].volume[vn].id)=0 then
						if ext("open reg 0 0:" # WINDIR # "\system32\config\SYSTEM")=0 then
							// close the registry
							ext("close reg 0")
							// save information
							count=count+1
							lineinfo[count]=BuildWinPartInfo(hd, h.partition[pn].volume[vn], h.secsize)
							// add it to the list
							listbuf=listbuf # lineinfo[count].desc # "^n"
						end if
						// close the drive
						ext("umount 0:")
					end if
					// next volume
					vn=vn+1
				wend

				// next partition
				pn=pn+1
			wend
		end if
		if VirtWinDrive<>"" then
			ExitLoop
		end if
	// Next hd
	next

	// Disable drive letters
	GetDrvLtrInfo(0)

	winsel=0
	if listbuf<>"" then
		while winsel=0
			// ask user for selection
			sel=GetListSelection("Select Windows Installation", listbuf, 1, "")
			if sel>0 then
				// use drive letter or mount as drive 0:
				if lineinfo[sel].dl<>"" then
					// use drive letter
					DrvLtr=lineinfo[sel].dl # ":"
				else
					// enable normal locking (to ensure 0: can be updated safely)
					ext("set option locking 1")
					if ext("mount 0: " # lineinfo[sel].hd # " " # lineinfo[sel].id)=0 then
						DrvLtr="0:"
					end if
				end if

				// ensure we have mounted drive available
				if DrvLtr<>"" then
					// check if we can access the System registry
					if ext("open reg 0 " # DrvLtr # WINDIR # "\system32\config\SYSTEM -f")=0 then
						// mount the root key as 0
						if ext("open key 0 ^"^" 0")=0 then
							// open CurrentControlSet
							if ext("open keyccs 5 0")=0 then
								// open the Software registry
								if ext("open reg 1 " # DrvLtr # WINDIR # "\system32\config\SOFTWARE -f")=0 then
									// mount the root key as 1
									if ext("open key 1 ^"^" 1")=0 then
										// success
										winsel=lineinfo[sel]
										winsel.DrvLtr=DrvLtr
									end if
								end if
							end if
						end if
					end if
					if winsel=0 then
						// clean up
						if DrvLtr="0:" then
							ext("umount 0:")
						else
							ext("close reg 0")
							ext("close reg 1")
						end if
					end if
				end if

				// notify of general failure
				if winsel=0 then
					GetListSelection("Error", "  Unable to access the Windows installation  ^n", 1, "err")
				end if
			else
				// user pressed escape
				ExitLoop
			end if
		wend

	elseif pn=0 and IsElevated()=0 then
		if GetListSelection("No Windows Installation found", "Restart script as Administrator^nContinue running as current user^n", 1, "warn") = 1 then
			ext("elevate ^"" # arg(0) # "^"")
			// clean-up screen of any error messages if user cancel when elevating privileges
			DrawTitle(0)
		end if

	else
		// Return to main menu if no Windows installation found
		GetListSelection("", "  No Windows Installation found  ^n", 1, "warn")
	end if

	// turn back on warning/error messages
	ext("set option quiet 0")

	// return result to caller
	return winsel
end sub


//-------------------------------------------------------------------------
// Purpose: Build partition informations of a Windows installation
//
// Input:   hd        - [i] hd number of the partition
//          part      - [i] partition informations
//          secsize   - [i] size of a sector in bytes
//          windir    - [i] windows directory
//
// Output:  1 with members:
//              .hd        - hd number
//              .id        - partition id
//              .dl        - drive letter of the partition
//              .name      - Name of partition or volume
//              .size      - Size of partition or volume
//              .desc      - description of the partition
//
// Notes:   
//
sub BuildWinPartInfo(hd, part, secsize)
	r=1
	r.hd=hd
	r.id=part.id
	r.name=part.name
	r.size=part.size

	// Build description
	if part.id<0x100 then
		type="Partition"
		n=2
	else
		type="Volume "
		n=4
	end if

	// length of the name part
	l=20
	if IsStrType(hd, 0) then
		r.desc="HD" # hd # ": "
		l=21-len(hd)-5*(tbsenv="WINDOWS" or tbsenv="UEFI")
	end if

	r.desc=r.desc # PAD(Left(part.name,l,0), l, 0, 0) # " " # type # " (" # Right("0000" # hex(part.id),n) # ") " # PAD("(" # Size2Text(part.size/(0x100000/secsize), 2) # ")", 13, 2) # " "
	// add fsname and drive letter
	if part.drvltr<>"" then
		r.desc=r.desc # PAD(part.fsname, 17) # " (" # part.drvltr # ":)"
		r.dl=part.drvltr
	else
		r.desc=r.desc # part.fsname
	end if

	return r
end sub


//-------------------------------------------------------------------------
// Purpose: Get selection of a file
//
// Input:   title   - [i] title for window
//          mask    - [i] search mask (don't include path - empty to explore all)
//                        use ";" as separator to search many masks: "*.vhd;*.vmdk;*.vhdx"
//          path    - [i] default path
//
// Output:  full path of file selected
//
// Notes:   Hidden and system directories and files are ignored
//
sub GetFileSelection(title, mask, path)

	if (tbsenv="WINDOWS" or tbsenv="UEFI") and gDrive=0 then
		gDrive=1
		// process all drives to get descriptions of drive letters
		// enable drive letters
		GetDrvLtrInfo(1)
		for hd=0 to 63
			h=GetHdInfo(hd)
			if h then
				// process partitions
				pn=(h.partition[0]=0)
				while h.partition[pn]
					if h.partition[pn].drvltr<>"" then
						// Build description of drive letter
						gDrive[asc(h.partition[pn].drvltr)]=BuildDriveDesc(hd, h.name, h.partition[pn])
					end if
					// process volumes
					vn=1
					while h.partition[pn].volume[vn]
						if h.partition[pn].volume[vn].drvltr<>"" then
							// Build description of drive letter
							gDrive[asc(h.partition[pn].volume[vn].drvltr)]=BuildDriveDesc(hd, h.name, h.partition[pn].volume[vn])
						end if
						// next volume
						vn=vn+1
					wend
					// next partition
					pn=pn+1
				wend
			end if
		next
		// disable drive letters
		GetDrvLtrInfo(0)
	end if

	// setup default mask to browse all files:
	// Windows, Linux FS and mounted drive numbers under DOS accept both * and *.* masks
	// Mounted drive numbers under Linux accept only *
	// Drive letters under DOS accept only *.*
	// we'll use * by default, unless under DOS where *.* will be used for drive letters
	// setup default mask
	if mask="" then
		mask="*"
	end if

	// length of the time field
	lt=len(GetTime())

	// keep previous path or read current directory
	if gLastPath<>"" and path="" then
		path=gLastPath
	end if
	// check path still exists (unless for UNC)
	if path<>"" and Left(path, 2)<>"\\" then
		// for drive number, check if drive is not mounted anymore, and mount it if possible
		if IsStrType(Left(path,1), 0) and gDrive[asc(path)]<>"" then
			if IsDrive(path)=0 then
				if ext("mount " # Left(path, 2) # " " # gDrive[asc(path)].drive) then
					// save cannot be mounted anymore
					gDrive[asc(path)]=""
				end if
			end if
		end if
		if SetDrive(path)=0 or ChDir(path)>0 then
			// if SetDrive(path) had work but not ChDir(path), stay in the drive and explore its current dir
			// no need to return to initial drive
			// check with last path if not already the case
			if path<>gLastPath then
				path=gLastPath
			else
				path=""
			end if
			if path<>"" and Left(path, 2)<>"\\" then
				if SetDrive(path)=0 or ChDir(path)>0 then
					// explore current directory of drive set
					path=""
				end if
			end if
		end if
	end if
	if path="" then
		path=GetCWD()
	end if
	// save path at begining
	gLastPath=path

	while 1
		listbuf=""
		lineinfo=""
		count=0

		ti=TXGetInfo()
		TXSetAttr(COLOR_NORMAL)
		TXGotoXY(1, ti.ViewHeight)
		pathdisp=BuildPath(BuildPathDesc(path, ti.ViewWidth-12-len(mask)), mask)
		print(PAD(Left("Listing of " # pathdisp, ti.ViewWidth-1, 0), ti.ViewWidth-1, 1, 0))

		// Add header
		listbuf=listbuf # "--[Date]--|" # Replace(PAD("-[Time]-", lt, 1), " ", "-") # "|---[Size]---|-[File]^n"
		count=count+1
		lineinfo[count]=-1

		// build list of directories and files
		if mask="*" then
			Lmask=""
		elseif Left(mask, 2)<>"*." then
			Lmask=LCase(mask)
		else
			Lmask=LCase(mask) # ";"
		end if
		filebuf=""
		fileinfo=""
		dfltmask="*"
		if tbsenv="DOS" and IsStrType(Left(path,1), 2) then
			dfltmask="*.*"
		end if
		f=findfirst(BuildPath(path, dfltmask))
		while len(f)
			// ignore the "." entry, hidden and system files, and the ".." entry on the root
			if f.name="." or (f.attrib & FILEATTR_HID) or (f.attrib & FILEATTR_SYS) then
				// ignored
			elseif f.name=".." and (path="/" or Mid(path,2)=":\" or Mid(path,2)=":/") then
				// ignored
			// directory
			elseif (f.attrib & FILEATTR_DIR) then
				// output information to buffer
				listbuf=listbuf # PAD(f.date, 10, 2) # " " # PAD(f.time, lt, 2)
				listbuf=listbuf # PAD("  <DIR>", 13)
				listbuf=listbuf # " " # f.name # "^n"
				// save current information
				count=count+1
				lineinfo[count].name=f.name
				lineinfo[count].attr=f.attrib
			// file: check if fits
			elseif InStr(Lmask, "*." # LCase(GetFileExtension(f.name)) # ";") or LCase(f.name)=Lmask or Lmask="" then
				// fit: save info in temporary buffer
				filebuf=filebuf # PAD(f.date, 10, 2) # " " # PAD(f.time, lt, 2)
				filebuf=filebuf # PAD(Size2Text(f.size, 0), 13, 2)
				filebuf=filebuf # " " # f.name # "^n"
				fileinfo.count=fileinfo.count+1
				fileinfo[fileinfo.count].name=f.name
				fileinfo[fileinfo.count].attr=f.attrib
			end if
			// next
			f=findnext(f)
		wend
		// add files to buffer list
		listbuf=listbuf # filebuf
		for i=1 to fileinfo.count
			count=count+1
			lineinfo[count]=fileinfo[i]
		next

		// add drive letters to the mix with attr = -1
		listbuf=listbuf # "--[Drives]--^n"
		count=count+1
		lineinfo[count]=-1
		for i=0 to 35
			// build drive letter (A to Z or 0 to 9)
			if i>25 then
				dl=chr(asc("0") + (i-26))
			else
				dl=chr(asc("A") + i)
			end if
			// check if drive available and is not current drive
			if IsDrive(dl) and dl<>UCase(Left(path, 1)) then
				if gDrive[asc(dl)].file=1 then
					listbuf=listbuf # " " # dl # ": " # BuildPathDesc(gDrive[asc(dl)], ti.ViewWidth-14) # "^n"
				else
					listbuf=listbuf # " " # dl # ": " # gDrive[asc(dl)] # "^n"
				end if
				// save current information
				count=count+1
				lineinfo[count].name=dl
				lineinfo[count].attr=-1
			end if
		next

		// add option to revert back to root file system under Linux
		if tbsenv="LINUX" and Left(path, 1)<>"/" then
			listbuf=listbuf # " {Linux File System}^n"
			count=count+1
			lineinfo[count].name=""
			lineinfo[count].attr=-1
		end if

		// add option to mount partition or volume
		if IsStrType(Left(VirtWinDrive,1), 0) then
			// Virtual hd is inside a numeric drive. Offer possibility to mount next drive
			drvnum=Left(VirtWinDrive, 1)+1
		else
			drvnum=1
		end if
		listbuf=listbuf # "Mount a Partition as Drive " # drvnum # ":^n"
		count=count+1
		lineinfo[count].attr=-3

		// mounting CD/DVD under DOS or TBOS
		if tbsenv="DOS" or tbsenv="UEFI" then
			// add option to mount CD/DVD
			listbuf=listbuf # "Mount CD/DVD (ISO9660) as Drive " # drvnum # ":^n"
			count=count+1
			lineinfo[count].attr=-4
		end if

		// add separator
		listbuf=listbuf # "--[Options]--^n"
		count=count+1
		lineinfo[count]=-1

		// add manual input option (feel free to add an input field handler)
		listbuf=listbuf # "Manually Enter Path (such as UNC)^n"
		count=count+1
		lineinfo[count].attr=-2

		if orgmask="" then
			// add option to mount virtual drive
			listbuf=listbuf # "Mount a Virtual Floppy/ISO file^n"
			count=count+1
			lineinfo[count].attr=-5
		end if

		sel=2
askfile:
		// prompt user for selection
		sel=GetListSelection(title, listbuf, sel, "")

		// title selected?
		if lineinfo[sel]=-1 then
			goto askfile

		// Mounting virtual drive?
		elseif lineinfo[sel].attr=-5 then
			orgmask=mask
			mask=dfltmask
			orgtitle=title
			title="Select Virtual Floppy to Mount"

		// CD mount request?
		elseif lineinfo[sel].attr=-4 then
			MountCDDrive(drvnum)
			path=GetCWD()

		// mount request?
		elseif lineinfo[sel].attr=-3 then
			// mount drive 1:
			MountDriveLtr(drvnum)
			path=GetCWD()

		// manual input?
		elseif lineinfo[sel].attr=-2 then
			// make a fake window
			winpos=DrawWindow("", -1, -1, 70, 5, "")
			if winpos.x>0 then
				linepos=DrawHelpLine("", 5, "")
				TXSetAttr(COLOR_WINDOWTEXT)
				TXGotoXY(winpos.x+2, winpos.y+2)
				print("Enter Path: ")
				TXSetAttr(COLOR_INPUT)
				temp=GetStr("", 50)
				// remove help line
				TXPutBlock(linepos.block, linepos.x, linepos.y)
				linepos=""
				TXPutBlock(winpos.block, winpos.x, winpos.y)
				winpos=""
				if temp<>"" then
					// ensure correct separators
					temp=BuildPath(temp, "")
					if Mid(temp, 2, 1)=":" or (tbsenv="LINUX" and Left(temp, 1)="/") then
						// full path: ensure drive exists
						temp=UCase(Left(temp,1)) # Mid(temp, 2)
						if SetDrive(temp)=0 then
							GetListSelection("", "  Drive " # Left(temp,2) # " not available  ^n", 1, "err")
							goto askfile
						end if
					elseif Left(temp, 1)<>"\" then
						temp=BuildPath(path, temp)
					end if
					// Ensure path exists
					if ChDir(temp)=0 then
						if Left(temp, 2)="\\" then
							path=temp
						else
							path=GetCWD()
						end if
					else
						// return back to previous path
						SetDrive(path)
						ChDir(path)
						GetListSelection(temp, "  Path not available  ^n", 1, "err")
						goto askfile
					end if
				else
					goto askfile
				end if
			end if

		// change drive?
		elseif lineinfo[sel].attr=-1 then
			if SetDrive(lineinfo[sel].name) then
				path=GetCWD()
			else
				GetListSelection("", "  Unable to open drive " # lineinfo[sel].name # ":  ^n", 1, "err")
				goto askfile
			end if

		// change dir?
		elseif (lineinfo[sel].attr & FILEATTR_DIR)<>0 then
			if sel.key=KEY_PROPERTY and mask=" " then
				ExitLoop
			// change to directory selected
			elseif ChDir(BuildPath(path, lineinfo[sel].name))=0 then
				// not using getcwd because path may contain manually entered UNC
				if lineinfo[sel].name=".." then
					path=RemoveComponent(path)
				else
					path=BuildPath(path, lineinfo[sel].name)
				end if
			else
				GetListSelection("", "  Unable to open directory " # lineinfo[sel].name # "  ^n", 1, "err")
				goto askfile
			end if

		// cancel mounting of floppy drive?
		elseif orgmask<>"" and sel<=0 then
			mask=orgmask
			orgmask=""
			title=orgtitle
			orgtitle=""

		// mount floppy drive?
		elseif orgmask<>"" then
			// Ensure drive number to mount is not the current drive
			if IsStrType(Left(path,1), 0) then
				// Mount next drive
				drvnum=Left(path,1)+1
			end if
			// Disable any warning / error messages
			ext("set option quiet 1")
			// Ensure drvnum is not already mounted
			ext("umount " # drvnum # ":")
			// Try to mount drive
			if ext("mount " # drvnum # ": ^"" # BuildPath(path, lineinfo[sel].name) # "^"")=0 then
				gDrive[asc(drvnum)]=BuildPath(path, lineinfo[sel].name)
				gDrive[asc(drvnum)].file=1
				// Return back to first selection
				mask=orgmask
				orgmask=""
				title=orgtitle
				orgtitle=""
				// Change current drive to the one mounted
				SetDrive(drvnum)
				path=GetCWD()
			else
				// Mount failed
				GetListSelection("","  Unable to mount " # lineinfo[sel].name # "  ^n", 1, "err")
				ext("set option quiet 0")
				goto askfile
			end if
			// Enable any warning / error messages
			ext("set option quiet 0")

		else
			ExitLoop
		end if
	wend

	// get rid of line we put on bottom
	TXSetAttr(COLOR_NORMAL)
	TXGotoXY(1, ti.ViewHeight)
	print(PAD("", ti.ViewWidth-1))

	// return nothing if escape
	if sel<=0 then
		return ""
	end if

	// return user selection
	if path="" then
		path=GetCWD()
	end if
	// save current directory
	gLastPath=path

	path=BuildPath(path, lineinfo[sel].name)
	return path
end sub

//-------------------------------------------------------------------------
// Purpose: Build description of a path
//
// Input:   path   -     complete path
//          maxlen - [i] Max target length
//
// Output:  Path truncated to fit the max length
//
// Notes:   each line item in the input list is terminated using a newline.
//
sub BuildPathDesc(path, maxlen)
	// Truncate any trailing separator
	if len(path)>1 and (Right(path,1)="\" or Right(path,1)="/") then
		path=Left(path, len(path)-1)
	end if

	if len(path, 0)>maxlen then
		// Split path to get last part
		f=SplitPath(path)
		// min length for path (included last separator)
		if len(f.path, 0)<=3 then
			minlp=len(f.path, 0)
		else
			minlp=7
		end if
		// Truncate the last part if needed
		if len(f.file, 0)>maxlen-minlp then
			// extract extension
			ext=GetFileExtension(f.file)
			f.file=Left(f.file, maxlen-minlp-len(ext)-2, 0) # ".." # ext
		end if
		// Truncate path if needed
		if len(f.path, 0)>maxlen-len(f.file, 0) then
			f.path=Left(f.path, maxlen-len(f.file, 0)-4, 0) # "..."
		end if
		// Build path
		path=BuildPath(f.path, f.file)
	end if

	// return description to caller
	return path
end sub

//-------------------------------------------------------------------------
// Purpose: Mount CD/DVD Drive under DOS or TBOS
//
// Input:   drvnum - [i] Drive number to mount.
//
// Output:  1 on success
//
// Notes:   Only the first caracter of drvnum is read.
//          Under LINUX, drvnum is not used
//
sub MountCDDrive(drvnum)
	drvnum=Left(drvnum, 1) # ":"
	listbuf=""
	subitems=""
	result=0

	if tbsenv="DOS" or tbsenv="UEFI" then
		// Check if drive already mounted
		if IsDrive(drvnum) then
			if GetListSelection("Drive " # drvnum # " is already mounted", "Unmount it^nCancel operation^n", 1, "warn")=1 then
				// dismount drive drvnum
				ext("umount " # drvnum)
			end if
		end if

		if IsDrive(drvnum)=0 then
			// Wait for the user to insert disk before creating the list of available drive
			if GetListSelection("", "Insert disc in a drive and press <Enter> when ready...^nCancel operation^n", 1, "warn")=1 then
				// Create list of available optical drives
				lineinfo=""
				count=0
				for n=0 to 3
					h=GetHdInfo("o" # n)
					if h then
						// save current information
						count=count+1
						lineinfo[count]="o" # n
						lineinfo[count].name=h.name
						listbuf=listbuf # "#" # n # " - " # h.name # "^n"
					end if
				next
				// Create list of available optical drives in USB if no other cd drives found
				if count=0 and tbsenv="DOS" then

lstusbcd:
					// Prevent lockup when accessing certain USB controllers
					ext("set option usblio 1")
					for n=0 to 3
						h=GetHdInfo("uo" # n)
						if h then
							// save current information
							count=count+1
							lineinfo[count]="uo" # n
							lineinfo[count].name=h.name
							listbuf=listbuf # "#u" # n # " - " # h.name # "^n"
							subitems=""
						end if
					next
				end if
				if count=0 then
					GetListSelection("", "    No CD/DVD drive found!    ^n", 1, "err")
				elseif subitems<>"" then
					GetListSelection("", "    No USB CD/DVD drive found!    ^n", 1, "err")
					subitems=""
				elseif tbsenv="DOS" then
					// add option to list manually usb drives
					subitems="Access USD optical drives^n"
				end if
			end if
		end if
	end if

	if listbuf<>"" then
		title="CD/DVD drive found"
		sel=GetListSelection(title, listbuf # subitems, 1, "")
		while sel>=1
			if sel=count+1 and subitems<>"" then
				// build list of USB cd drives
				ExitLoop

			else
				// disable various warning / error messages
				ext("set option quiet 1")
				// try mounting this one
				if ext("mount " # drvnum # " " # lineinfo[sel])=0 then
					result=1
					gDrive[asc(drvnum)]="(" # lineinfo[sel].name # ")"
					// Set current drive to CD on manual mode
					if mode="manu" then
						SetDrive(drvnum)
					end if
					ExitLoop
				else
					// Warn user it didn't mount
					GetListSelection("", "    Mount of drive " # lineinfo[sel] # " failed!    ^n", 1, "err")
				end if
			end if
			// Mount failed so ask again the drive to mount
			sel=GetListSelection("Select drive to mount", listbuf # subitems, sel, "")
		wend

		// Enable various warning / error messages
		ext("set option quiet 0")
		if sel=count+1 and subitems<>"" and result=0 then
			// build list of USB cd drives
			goto lstusbcd
		end if
	end if

	return result
end sub

//-------------------------------------------------------------------------
// Purpose: Mount Drive to a partition/volume
//
// Input:   drvnum - [i] Drive number to mount. If empty, 1: is used by default.
//
// Output:  na
//
// Notes:   Only the first caracter of drvnum is read.
//
sub MountDriveLtr(drvnum)
	drvnum=Left(drvnum, 1) # ":"

	// enable drive letters
	GetDrvLtrInfo(1)

	// display menu of hard drives available
	listbuf=""
	lineinfo=""
	count=0

	for hd=0 to 63
		h=GetHdInfo(hd)
		if h then
			// count line
			count=count+1
			lineinfo[count]=""
			// output label
			if h.gpt=1 then
				type="GPT"
			elseif h.embr=1 then
				type="EMBR"
			elseif h.partition[0]=0 then
				type="MBR"
			else
				type="VOL"
			end if
			listbuf=listbuf # "HD" # h.num # " - " # h.name # " - " # type # " (" # h.size/(0x100000/h.secsize) # " MiB)^n"
			// add partitions/volumes
			pn=(h.partition[0]=0)
			while h.partition[pn]
				// save line information
				count=count+1
				lineinfo[count]=BuildPartInfo(hd, h.partition[pn], h.secsize)
				lineinfo[count].hdname=h.name
				listbuf=listbuf # lineinfo[count] # "^n"
				// process volumes
				vn=1
				while h.partition[pn].volume[vn]
					// save line information
					count=count+1
					lineinfo[count]=BuildPartInfo(hd, h.partition[pn].volume[vn], h.secsize)
					lineinfo[count].hdname=h.name
					listbuf=listbuf # lineinfo[count] # "^n"
					// next volume
					vn=vn+1
				wend
				// next partition
				pn=pn+1
			wend
		end if
	next

	// Disable drive letters
	GetDrvLtrInfo(0)
	// disable various warning / error messages
	ext("set option quiet 1")

	if count>0 then
		// prompt user for selection
		sel=GetListSelection("Select Partition to mount", listbuf, 1, "")
		// determine if user selected something
		while sel>0
			if lineinfo[sel]="" then
				// full disk selected
			elseif lineinfo[sel].dl="" then
				// dismount any drive one now
				ext("umount " # drvnum)
				// try mounting this one
				res=ext("mount " # drvnum # " " # lineinfo[sel].hd # " " # lineinfo[sel].id)
				if res=0 then
					gDrive[asc(drvnum)]=BuildDriveDesc(lineinfo[sel].hd, lineinfo[sel].hdname, lineinfo[sel])
					SetDrive(drvnum)
					ExitLoop
				end if
				// warn user it didn't mount
				GetListSelection("", "    Mount failed!    ^n", 1, "err")
			elseif GetListSelection("Selected partition already available at drive " # lineinfo[sel].dl # ":", "Change drive to " # lineinfo[sel].dl # ":^nChoose another partition^n", 1, "warn")=1 then
				SetDrive(lineinfo[sel].dl)
				ExitLoop
			end if
			sel=GetListSelection("Select Partition", listbuf, sel, "")
		wend

	elseif pn=0 and IsElevated()=0 then
		if GetListSelection("Elevate Privileges", "Restart script as Administrator^nContinue running as current user^n", 1, "warn") = 1 then
			ext("elevate ^"" # arg(0) # "^"")
			// clean-up screen of any error messages if user cancel when elevating privileges
			DrawTitle(0)
		end if

	elseif pn=0 then
		GetListSelection("", "  No HD drive found  ^n", 1, "warn")

	else
		GetListSelection("", "  No available partition found  ^n", 1, "err")
	end if

	// enabled various warning / error messages
	ext("set option quiet 0")
end sub

//-------------------------------------------------------------------------
// Purpose: Build description of a drive letter for physical drives
//
// Input:   hd   - [i] number of hd
//          name - [i] name of hd
//          part - [i] Informations about the partititon or volumne (.name and .id)
//
// Output:  Description of the drive letter, with .drive set to the information needed to mount the drive
//
// Notes:
//
sub BuildDriveDesc(hd, name, part)
	if part.id=0 then
		type="Volume"
		n=0
	elseif part.id<0x100 then
		type="Partition"
		n=2
	else
		type="Volume"
		n=4
	end if
	drive=hd # " 0x" # hex(part.id)

	// Build description
	if IsStrType(hd, 0) then
		if name<>"" then
			name=" (" # name # ")"
		end if
		hd="HD" # hd # name
	end if
	desc="(" # part.name # ") - " # hd # " " # type # " (" # Right("0000" # hex(part.id), n) # ")"
	desc.drive=drive

	return desc
end sub

//-------------------------------------------------------------------------
// Purpose: Build partition informations
//
// Input:   hd        - [i] hd number of the partition
//          part      - [i] partition information
//          secsize   - [i] size of a sector in bytes
//
// Output:  Description of the partition with members:
//              .hd      - hd number
//              .id      - partition id
//              .dl      - drive letter of the partition
//              .name    - name of partition
//
// Notes:   
//
sub BuildPartInfo(hd, part, secsize)
	if part.id=0 then
		type="Volume    "
		n=-1
	elseif part.id<0x100 then
		type="Partition"
		n=0
	else
		type="Volume"
		n=2
	end if

	// Build description
	if part.drvltr<>"" then
		// Show drive letter to user
		r=" (" # part.drvltr # ":) "
	elseif tbsenv="WINDOWS" or tbsenv="UEFI" then
		// Physical partition under Windows
		r="      "
	else
		// Physical partition under DOS or LINUX
		r="  "
	end if

	// nbr of sectors per MiB
	sectpermib=0x100000/secsize
	// width for the part name
	w=24-len(r)
	r=r # PAD(Left(Left("  ",n) # part.name, w, 0), w, 0, 0) # PAD("(" # (part.size+sectpermib/2)/sectpermib # " MiB)", 15, 2) # " " # PAD(type, 9, 0) # " (" # Right("0" # hex(part.id), 2+n) # ") " # Left("  ", 2-n) # part.fsname

	// Add properties
	r.hd=hd
	r.id=part.id
	r.dl=part.drvltr
	r.name=part.name

	// return result
	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Combine a path and file
//
// Input:   path - file path
//          file - file name
//
// Output:  combined path and file
//
// Notes:   used to handle trailing separator
//
sub BuildPath(path, file)
	// clean up separators
	if tbsenv="LINUX" then
		sepsch="\"
		seprpl="/"
	else
		sepsch="/"
		seprpl="\"
	end if

	pos=InStr(path, sepsch)
	while pos>0
		path=Left(path, pos-1) # seprpl # Mid(path, pos+1)
		pos=InStr(path, sepsch)
	wend

	// add separator if needed
	if len(path)>0 and Right(path, 1)<>seprpl then
		path=path # seprpl
	end if

	// ensure file doesn't have separator
	if Left(file, 1)=seprpl or Left(file, 1)=sepsch then
		file=Mid(file, 2)
	end if
	pos=InStr(file, sepsch)
	while pos>0
		file=Left(file, pos-1) # seprpl # Mid(file, pos+1)
		pos=InStr(file, sepsch)
	wend

	// finish combine
	path=path # file
	// return path
	return path
end sub

//-------------------------------------------------------------------------
// Purpose: Remove list component of file path
//
// Input:   filepath  - file path to process
//
// Output:  truncated file path
//
// Notes:
//
sub RemoveComponent(filepath)
	// split file path
	s=SplitPath(filepath)
	// return path
	return s.path
end sub

//-------------------------------------------------------------------------
// Purpose: Split a file in to its path / file components
//
// Input:   filepath - path to split
//
// Output:  split path with members .path and .file set
//
// Notes:   .path may include a trialing separator
//
sub SplitPath(filepath)
	// search for separator from end to beginning
	pos=len(filepath)-1
	char=Mid(filepath, pos, 1)
	while pos>0 and char<>"/" and char<>"\"
		pos=pos-1
		char=Mid(filepath, pos, 1)
	wend

	// extract the parts
	r.path=Left(filepath, pos)
	r.file=Mid(filepath, pos+1)

	// Return result to caller
	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Read file component of file path
//
// Input:   filepath  - file path to process
//
// Output:  truncated file name
//
// Notes:
//
sub GetComponent(filepath)
	// split file path
	s=SplitPath(filepath)
	// return file name
	return s.file
end sub

//-------------------------------------------------------------------------
// Purpose: Get the extension of a file
//
// Input:   filename - [i] File name
//
// Output:  extension of the file (without the dot)
//
// Notes:
//
sub GetFileExtension(filename)
	// default: no extension
	ext=""

	// init position of the last dot
	pos=len(filename)-1
	// loop from end until dot found
	while pos>1
		if Mid(filename, pos, 1)="." then
			ext=Mid(filename, pos+1)
			ExitLoop
		end if
		pos=pos-1
	wend

	// return file extension
	return ext
end sub

//-------------------------------------------------------------------------
// Purpose: Replace a char inside a string to another char
//
// Input:   s      -     String to process
//          char1  - [i] char to be replaced
//          char2  - [i] char to replace
//
// Output:  Cleaned string
//
// Notes:   char1 has to be non-empty
//
sub Replace(s, char1, char2)
	// Get length of char to be replaced
	l=len(char1)
	// Init result
	r=""

	// Get position of char to replace in the string
	pos=InStr(s, char1)
	// Loop until no more char to replace
	while pos>0
		r=r # Left(s, pos-1) # char2
		// next part
		s=Mid(s, pos+l)
		pos=InStr(s, char1)
	wend
	// add last part
	r=r # s

	return r
end sub

//-------------------------------------------------------------------------
// Purpose: Remove leading and trailing spaces, tabs, newlines and all
//          non-readable char from a string
//
// Input:   s  - string to process
//
// Output:  Trim string
//
// Notes:
//
sub Trim(s)
	l=len(s)
	i=1
	// Remove leading spaces and all non-readable chars
	while Mid(s, i, 1)<=" " and i<=l
		i=i+1
	wend
	// Remove trailing spaces and all non-readable chars
	while Mid(s, l, 1)<=" " and i<l
		l=l-1
	wend
	// return trim string
	return Mid(s, i, l-i+1)
end sub

//-------------------------------------------------------------------------
// Purpose: Open disk cache on a disk (WINDOWS or DOS only)
//
// Input:   hdnum   - [i] Number of hd
//
// Output:  None
//
// Notes:   global gUseHDCache must be set to 1
//
sub OpenDiskCache(hdnum)
	if tbsenv="LINUX" then
		// cache cannot be open under Linux
	elseif gUseHDCache=1 and gUseHDCache[hdnum]=0 then
		print("Opening disk cache on HD", hdnum, "...")
		// option for Lazy writing
		option=""
		if HDLAZY_WRITE=1 then
			option=" /p"
		end if

		// open cache
		if ext("open cache " # hdnum # " " # HDCACHE_SIZE # " /l=" # HDCACHE_IOSIZE # option)=0 then
			// save cache is enable
			gUseHDCache[hdnum]=1
			printl("Complete")
		else
			// save to not try to enable cache again
			gUseHDCache[hdnum]=-1
			printl("Failed")
		end if
	end if
end sub

//-------------------------------------------------------------------------
// Purpose: Closes disk cache on a drive
//
// Input:   hdnum   - [i] number of HD
//          flush   - [i] 1 to flush cache only, 0 to close cache
//
// Output:  1 on success or cache already closed, 0 on failure
//
// Notes:   gUseHDCache[hdnum] must be 1
//
sub CloseDiskCache(hdnum, flush)
	if gUseHDCache[hdnum]=1 then
		option=""
		if flush=1 then
			print("Flushing disk cache on HD", hdnum, "...")
			option=" /w"
		else
			print("Closing disk cache on HD", hdnum, "...")
		end if

		// close cache
		if ext("close cache " # hdnum # option)=0 then
			printl("Complete")
			if flush=0 then
				// save cache is disable
				gUseHDCache[hdnum]=0
			end if
		else
			printl("Failed")
			printl("  - Disk Write Failure - Registry and File System Data Loss Likely")
			// force close cache
			ext("close cache " # hdnum # " /f")
			// save to not try to enable cache again
			gUseHDCache[hdnum]=-1
		end if
	end if
end sub

//-------------------------------------------------------------------------
// Purpose: Evaluate a size in numeric, to a size with unit (B, KiB, MiB, GiB, TiB)
//
// Input:   size   -     size
//          i      - [i] 0 if the entry size is in bytes, 1 if it's KiB, 2 if it's MiB...
//
// Output:  Size in B, KiB, MiB, GiB or TiB, rounded at 2 digits after coma
//
// Notes:
//
sub Size2Text(size, i)
	r=0
	while size>9999 and i<4
		r=size%1024
		size=size/1024
		// determine the part after coma
		r=r*1000/1024+5
		if r>=1000 then
			size=size+1
			r=1
		end if
		// number of loop done
		i=i+1
	wend
	// determine the unit
	if i=0 then
		unit="B  "
	elseif i=1 then
		unit="KiB"
	elseif i=2 then
		unit="MiB"
	elseif i=3 then
		unit="GiB"
	else
		unit="TiB"
	end if
	// add the coma and the unit
	if r>0 then
		sep="."
		if IsStrType("0,1", 1) then
			// decimal separator is the coma
			sep=","
		end if
		size=size # sep # Left(Right("000" # r, 3), 2)
	end if
	size=size # " " # unit
	// return result
	return size
end sub
